/* vi:set ts=4 sw=4:
=======================================================================

  fcstring version 0.0.0

  Copyright(c) 1999 Minero Aoki
  aamine@dp.u-netsurf.ne.jp


  This program is Free Software.
  You can distribute it and/or modify it under the terms of the "GPL2",
  GNU General Public License version 2 as published by the Free Software
  Foundation.

  This program is distributed in the hope that it will be usefull,
  but WITHOUT ANY WARRANTY, without even the implied warranty of
  merchantabiliry or fitness for a particular purpose.  See the file
  "GPL-2" for more details.

=======================================================================
*/


#include <stdio.h>
#include <string.h>
#ifdef __STDC__
#  include <stdlib.h>
#endif
#include "fcstring.h"


/* -------------------- class fcchar --------------------- */


int
fcc_isnl(ch)
    fcchar ch;
{
    return FCC_ISNL(ch);
}


int
fcc_isalpha(ch)
    fcchar ch;
{
    return FCC_ISALPHA(ch);
}


int
fcc_islower(ch)
    fcchar ch;
{
    return FCC_ISLOWER(ch);
}


int
fcc_isupper(ch)
    fcchar ch;
{
    return FCC_ISUPPER(ch);
}


int
fcc_isdigit(ch)
    fcchar ch;
{
    return FCC_ISDIGIT(ch);
}


int
fcc_isalnum(ch)
    fcchar ch;
{
    return FCC_ISALNUM(ch);
}


int
fcc_iscntrl(ch)
    fcchar ch;
{
    return FCC_ISCNTRL(ch);
}


int
fcc_isspace(ch)
    fcchar ch;
{
    return FCC_ISSPACE(ch);
}


int
fcc_isgraph(ch)
    fcchar ch;
{
    return FCC_ISGRAPH(ch);
}


int
fcc_isprint(ch)
    fcchar ch;
{
    return FCC_ISPRINT(ch);
}


int
fcc_ispunct(ch)
    fcchar ch;
{
    return FCC_ISPUNCT(ch);
}


int
fcc_isblank(ch)
    fcchar ch;
{
    return FCC_ISBLANK(ch);
}


int
fcc_isxdigit(ch)
    fcchar ch;
{
    return FCC_ISXDIGIT(ch);
}


fcchar
fcc_tolower(ch)
    fcchar ch;
{
    return FCC_TOLOWER(ch);
}


fcchar
fcc_toupper(ch)
    fcchar ch;
{
    return FCC_TOUPPER(ch);
}


/* ===================================================================
                         class fcstring
====================================================================== */


#define  fcclen2byte(len)  (sizeof(fcchar) * (len))
#define  fcca_cmp(dest,src,len)  memcmp( dest, src, fcclen2byte(len) )
#define  fcca_cpy(dest,src,len)  memmove( dest, src, fcclen2byte(len) )

#ifndef max
#  define max(a,b)  ((a) > (b) ? (a) : (b))
#  define min(a,b)  ((a) > (b) ? (b) : (a))
#endif

#define FCS_NEW_ALLOC(var, len, retval) do {\
    var = fcs_new() ;\
    if (var == 0) return retval ;\
    if (! fcs_alloc(var, len)) return retval;\
} while(0)

#define FCS_ALLOC(str, len, retval) \
    if (! fcs_alloc(str, len)) return retval




/* allocate memory for fcstring */

static int
fcs_alloc(str, len)
    fcstring *str;
    int len;
{
    fcchar *tmp = 0;

    if (str->ptr != 0)
        tmp = realloc(str->ptr, fcclen2byte(len));
    else
        tmp = malloc(fcclen2byte(len));
    
    if (tmp == 0) return 0;
    str->ptr = tmp;

    return 1;
}


/* make new fcstring */

fcstring*
fcs_new()
{
    fcstring *ret;

    ret = (fcstring*)malloc(sizeof(fcstring));
    if (ret == NULL) return 0;

    ret->ptr = 0;
    ret->len = 0;

    return ret;
}


/* free fcstring */

void
fcs_free(str)
    fcstring *str;
{
    if (str->ptr) {
        free(str->ptr);
        str->ptr = 0;
    }
    free(str);
}


/* tell if a == b */

int
fcs_equal(a, b)
    fcstring *a, *b;
{
    if (a->len != b->len) return 0;
    
    if (a == b) return 1;
    if (fcca_cmp(a->ptr, b->ptr, a->len) == 0)
        return 1;
    else
        return 0;
}


/* tell if a == b, but case insensitive */

int
fcs_cequal(a, b)
    fcstring *a, *b;
{
    if (a->len != b->len) return 0;

    if (a == b) return 1;
    if (fcs_ccompare(a, b) == 0) return 1;
    return 0;
}


/* comparison function for qsort */

int
fcs_compare(a, b)
    fcstring *a, *b;
{
    if (a == b) return 0;
    return fcca_cmp(a->ptr, b->ptr, min(a->len, b->len));
}


/* comparison function for qsort, but case insensitive */

int
fcs_ccompare(a, b)
    fcstring *a, *b;
{
    long limit, i;

    if (a == b) return 0;
    
    limit = min(a->len, b->len);
    for (i = 0; i < limit; i++) {
        if (! fcc_eq(a->ptr[i], b->ptr[i])) {
            fcchar lowa = fcc_tolower(a->ptr[i]);
            fcchar lowb = fcc_tolower(b->ptr[i]);

            if (fcc_eq(lowa, lowb)) continue;
            return fcc_gt(lowa, lowb) ? 1 : -1;
        }
    }
    if (a->len == b->len) return 0;
    return a->len > b->len ? -1 : 1;
}


/* get part of string */

fcstring*
fcs_substr(str, idx, len)
    fcstring *str;
    long idx, len;
{
    fcstring *ret = 0;

    if (idx >= str->len) return 0;
    len = min(idx + len, str->len) - idx;

    FCS_NEW_ALLOC(ret, len, 0);

    fcca_cpy(ret->ptr, str->ptr + idx, len);
    ret->len = len;

    return ret;
}


/* get "idx"th char. An index of the first element is 0 */

fcchar
fcs_entry(str, idx)
    fcstring *str;
    long idx;
{
    if (idx < 0 || idx >= str->len) return FCC_NIL;
    return str->ptr[idx];
}


/* duplicate string */

fcstring*
fcs_dup(src)
    fcstring *src;
{
    fcstring *ret = 0;

    FCS_NEW_ALLOC(ret, src->len, 0);

    fcca_cpy(ret->ptr, src->ptr, src->len);
    ret->len = src->len;
    
    return ret;
}


/* "a" + "b" */

fcstring*
fcs_plus(a, b)
    fcstring *a, *b;
{
    fcstring *ret = 0;

    FCS_NEW_ALLOC(ret, a->len + b->len, 0);

    fcca_cpy(ret->ptr, a->ptr, a->len);
    fcca_cpy(ret->ptr + a->len, b->ptr, b->len);
    ret->len = a->len + b->len;

    return ret;
}


/* "str" + "str" + "str" + .... */

fcstring*
fcs_times(str, times)
    fcstring *str;
    long times;
{
    fcstring *ret = 0;
    long i;

    if (times < 0) return 0;
    
    FCS_NEW_ALLOC(ret, str->len * times, 0);

    ret->len = str->len * times;
    for (i = 0; i < times; i++)
        fcs_set(ret, 0 + str->len * i, str);
    
    return ret;
}


/* "dest" += "src" */

int
fcs_concat(dest, src)
    fcstring *dest, *src;
{
    if (src->len == 0) return 1;
    
    FCS_ALLOC(dest, dest->len + src->len, 0);

    fcca_cpy(dest->ptr + dest->len, src->ptr, src->len);
    dest->len += src->len;

    return 1;
}


/* replace string with new value */

int
fcs_replace(dest, src)
    fcstring *dest, *src;
{
    if (src == dest) return 1;

    FCS_ALLOC(dest, src->len, 0);

    dest->len = src->len;
    fcca_cpy(dest->ptr, src->ptr, src->len);

    return 1;
}


/* trancate size of string to 0 */

int
fcs_clear(str)
    fcstring *str;
{
    FCS_ALLOC(str, 0, 0);
    str->len = 0;
    return 1;
}


/* overwrite "dest" from index "idx", with "src" */

int
fcs_set(dest, idx, src)
    fcstring *dest, *src;
    int idx;
{
    int needlen = 0;

    if (idx > dest->len)   /* not '>=' */
        return 0;

    needlen = idx + src->len;
    if (needlen > dest->len) FCS_ALLOC(dest, needlen, 0);

    dest->len = needlen;
    fcca_cpy(dest->ptr + idx, src->ptr, src->len);

    return 1;
}


/* set char. NOT insert but overwrite */

int
fcs_enter(str, idx, ch)
    fcstring *str;
    int idx;
    fcchar ch;
{
    if (idx < 0)         return 0;
    if (idx >= str->len) return 0;

    str->ptr[idx] = ch;
    return 1;
}


/* insert string */

int
fcs_insert(dest, idx, src)
    fcstring *dest, *src;
    int idx;
{
    if (idx < 0)         return 0;
    if (idx > dest->len) return 0;

    FCS_ALLOC(dest, dest->len + src->len, 0);

    dest->len += src->len;
    fcca_cpy(dest->ptr + idx + src->len,
             dest->ptr + idx, 
             dest->len - idx);
    fcca_cpy(dest->ptr + idx,
             src->ptr,
             src->len);
    
    return 1;
}


/* insert char */

int
fcs_insertc(dest, idx, ch)
    fcstring *dest;
    int idx;
    fcchar ch;
{
    if (idx < 0)         return 0;
    if (idx > dest->len) return 0;

    FCS_ALLOC(dest, dest->len + 1, 0);

    dest->len++;
    fcca_cpy(dest->ptr + idx + 1,
             dest->ptr + idx, 
             dest->len - idx);
    dest->ptr[idx] = ch;
    
    return 1;
}


/* delete string */

int
fcs_delete(str, idx, len)
    fcstring *str;
    int idx, len;
{
    if (idx < 0)         return 0;
    if (idx >= str->len) return 0;
    if (len < 0)         return 0;

    len = min(idx + len, str->len) - idx;
    fcca_cpy(str->ptr + idx,
             str->ptr + idx + len,
             str->len - len);
    FCS_ALLOC(str, str->len - len, 0);
    str->len -= len;

    return 1;
}


/* delete char */

int
fcs_deletec(str, idx)
    fcstring *str;
    int idx;
{
    if (idx < 0)        return 0;
    if (idx > str->len) return 0;

    fcca_cpy(str->ptr + idx,
             str->ptr + idx + 1,
             1);
    str->len--;
    
    return 1;
}


/* cut last char */

int
fcs_chop(str)
    fcstring *str;
{
    if (str->len <= 0) return 0;
    str->len--;
    return 1;
}


/* cut CR or LF or CRLF at the end of the string */

int
fcs_chomp(str)
    fcstring *str;
{
    if (str->len <= 0) return 0;
    if (FCC_ISNL(str->ptr[str->len - 1])) {
        str->len--;
        return 1;
    }
    return 0;
}


/* delete spaces at head/tail */

int
fcs_strip(str)
    fcstring *str;
{
    fcchar *p;

    for (p = str->ptr + (str->len - 1); p > str->ptr; p--) {
        if(! fcc_isspace(*p))
            break;
        else
            str->len--;
    }
    for (p = str->ptr; p < str->ptr + str->len; p++) {
        if(! fcc_isspace(*p))
            break;
    }
    if (p != str->ptr)
        fcs_cuth(str, p - str->ptr);
    
    return 1;
}


/* cut off string from tail */

int
fcs_cutt(str, len)
    fcstring *str;
    int len;
{
    if (len >= str->len)
        str->len = 0;
    else
        str->len -= len;
    
    return 1;
}


/* convert all lower chars to upper chars */

int
fcs_toupper(str)
    fcstring *str;
{
    int i;

    for (i = 0; i < str->len; i++)
        str->ptr[i] = FCC_TOUPPER(str->ptr[i]);
    return 1;
}


/* convert all upper case char to lower case char */

int
fcs_tolower(str)
    fcstring *str;
{
    int i;

    for (i = 0; i < str->len; i++)
        str->ptr[i] = FCC_TOLOWER(str->ptr[i]);
    return 1;
}


/* count existing char in string */

long
fcs_count(str, ch)
    fcstring *str;
    fcchar ch;
{
    fcchar *p    = str->ptr;
    fcchar *pend = str->ptr + str->len;
    long i = 0;

    for (; p < pend; p++)
        if (fcc_eq(*p, ch))
            i++;

    return i;
}


/* search char from head to tail */

long
fcs_findc(str, ch)
    fcstring *str;
    fcchar ch;
{
    fcchar *p = str->ptr;

    for (; p < str->ptr + str->len; p++) {
        if (fcc_eq(*p, ch))
            return p - str->ptr;
    }
    return FCS_NOTFOUND;
}


/* search char from tail to head */

long
fcs_rfindc(str, ch)
    fcstring *str;
    fcchar ch;
{
    fcchar *p = str->ptr + (str->len - 1);

    for (; p >= str->ptr; p--) {
        if (fcc_eq(*p, ch))
            return p - str->ptr;
    }
    return FCS_NOTFOUND;
}


/* search "targ" from "str" forward */

long
fcs_find(str, targ)
    fcstring *str, *targ;
{
    fcchar *p    = str->ptr;
    fcchar *pend = str->ptr + str->len - targ->len;

    if (targ->len > str->len) return FCS_NOTFOUND;
    if (targ->len == 0)       return FCS_NOTFOUND;

    for (; p <= pend; p++) {
        if (fcc_eq(*p, targ->ptr[0])) {
            if(fcca_cmp(p, targ->ptr, targ->len) == 0)
                return p - str->ptr;
        }
    }
    return FCS_NOTFOUND;
}


/* search "targ" from "str" backward */

long
fcs_rfind(str, targ)
    fcstring *str, *targ;
{
    fcchar *p    = str->ptr + str->len - targ->len;
    fcchar *pend = str->ptr;

    if (str->len < targ->len) return FCS_NOTFOUND;
    if (targ->len == 0)       return FCS_NOTFOUND;
    
    for (; p >= pend; p--) {
        if (fcc_eq(*p, targ->ptr[0]))
            if (fcca_cmp(p, targ->ptr, targ->len))
                return p - str->ptr;
    }
    return FCS_NOTFOUND;
}



/* ------------- functions for convertion --------------------- */

#define ASCII_ESC 0x1b

/* make char* with ISO-2022 encoding from fcstring */

char*
fcs_to_iso2022(str)
    fcstring *str;
{
    char    *arr;
    char    *dest;
    fcctype  ctype = FCC_TYPE_EN;
    fcctype  tmp;
    fcchar  *src;
    fcchar  *srcend;

/*
   temporary code !!!
*/
dest = arr = (char*)malloc(str->len * 4);

    src = str->ptr;
    srcend = src + str->len;

    for (; src < srcend && *src; src++) {
        if ((tmp = fcc_type(*src)) != ctype) {
printf("bug: code type changed: ctype=0x%x, prev=0x%x\n", tmp, ctype);
            ctype = tmp;
            *dest++ = ASCII_ESC;
        }
        if (fcc_issb(*src)) {
            *dest++ = fcc_a2(*src);
        } else {
            *dest++ = fcc_a1(*src);
            *dest++ = fcc_a2(*src);
        }
    }
    *dest = '\0';

    return arr;
}


/* make fcstring from out source (string is not duplicated) */

fcstring*
fcs_from_out(str, len)
{
    fcstring *ret;

    ret = fcs_new();
    if (ret == 0) return 0;

    ret->ptr = str;
    ret->len = len;

    return ret;
}


/* make fcstring from out source (string is duplicated) */

fcstring*
fcs_from_out_dup(str, len)
{
    fcstring *ret;

    FCS_NEW_ALLOC(ret, len, 0);
    fcca_cpy(ret->ptr, str, len);

    return ret;
}


/* make fcsting with char* string */

fcstring*
fcs_from_iso2022(arr, arrsize)
    char *arr;
    size_t arrsize;
{
    fcstring *ret;
    fcchar   *dest;
    char     *src = arr;
    char     *srcend = arr + arrsize;

#define writemb(ctype) do {\
    while (src < srcend && *src != ASCII_ESC) {\
        *dest = fcc_from2b(ctype, *src, *(src+1));\
        ret->len++;\
        dest++;\
        src += 2;\
    }\
} while (0)

#define writesb(ctype) do {\
    while (src < srcend && *src != ASCII_ESC) {\
        *dest = fcc_from1b(ctype, *src);\
        ret->len++;\
        dest++;\
        src++;\
    }\
} while (0)

    FCS_NEW_ALLOC(ret, arrsize, 0);   /* temp 256 fcchar */
    ret->len = 0;

    dest = ret->ptr;
    
    writesb(FCC_TYPE_EN);
    while (src < srcend) {
        src++;

        /* temporary code !!! */
        switch (*src) {
            /* \e$B  jisx0208 */
            case 0x24:
                src += 2;
                writemb(FCC_TYPE_JA);
                break;

            /* \e(B  ascii */
            case 0x28:
                src += 2;
                writesb( FCC_TYPE_EN );
                break;

            /* unknown seaquence */
            default:
                printf("fcs_from_iso2022: must not happen\n");
                exit(1);
                break;
        }
    }
    return ret ;
}


#define  EUCCHAR(ch) ((ch) & EUCMASK)
#define  EUCMASK     (1 << 7)
#define  EUCOFF(ch)  ((ch) & ~EUCMASK)

/* make new fcstring object from EUC-JP encoded char* */

fcstring*
fcs_from_eucjp(arr, arrsize)
    char *arr;
    size_t arrsize;
{
    fcstring *ret;
    fcchar *dest;
    char *p;
    char *pend;

    FCS_NEW_ALLOC(ret, 256, 0);
    ret->len = 0;

    p = arr;
    pend = p + arrsize;
    dest = ret->ptr;

    for (; p < pend; p++) {
        if (EUCCHAR(*p))
            *dest = fcc_from2b(FCC_TYPE_JA, EUCOFF(p[0]), EUCOFF(p[1]));
        else
            *dest = fcc_from1b(FCC_TYPE_EN, p[0]);
        dest++;
        ret->len++;
    }

    return ret;
}


#define  SJISCHAR(ch) ((ch) & SJISMASK)
#define  SJISMASK     (1 << 7)

/* make new fcstring object from shift_jis encoded char* */

fcstring*
fcs_from_sjis(arr, arrsize)
    char *arr;
    size_t arrsize;
{
    fcstring *ret;
    fcchar *dest;
    char *p;
    char *pend;

    FCS_NEW_ALLOC(ret, 256, 0);
    ret->len = 0;

    p = arr;
    pend = p + arrsize;
    dest = ret->ptr;

    for (; p < pend; p++) {
        if (SJISCHAR(*p))
            *dest = fcc_from2b(FCC_TYPE_JA, p[0] & ~SJISMASK, p[1]);
        else
            *dest = fcc_from1b(FCC_TYPE_EN, p[0]);
        dest++;
        ret->len++;
    }

    return ret;
}



/* ==================================================================
                         class fcbuffer
===================================================================== */


#define GAPPTR(buf)  ((buf)->ptr + (buf)->gap)
#define RIGHTPTR(buf)  ((buf)->ptr + (buf)->gap + (buf)->gapsize)
#define RIGHTIDX(buf)  ((buf)->gap + (buf)->gapsize)
#define BUFINITLEN 2048


static void
ungap_buffer(buf)
    fcbuffer *buf;
{
    if (buf->gapsize == 0) return;

    fcca_cpy(GAPPTR(buf),
             RIGHTPTR(buf),
             FCB_SIZE(buf) - RIGHTIDX(buf));
    buf->gap = buf->len;
    buf->gapsize = 0;
}


/* make new fcbuffer object */

fcbuffer*
fcb_new()
{
    fcbuffer *ret;

    ret = (fcbuffer*)malloc(sizeof(fcbuffer));
    if(ret == 0) return 0;

    ret->ptr = 0;
    ret->len = 0;
    ret->capa = 0;
    ret->gap = 0;
    ret->gapsize = 0;
    ret->mark = 0;
    ret->mark_end = 0;
    ret->mark_tbl = 0;

    ret->ptr = (fcchar*)malloc(fcclen2byte(BUFINITLEN));
    if (!ret->ptr) goto no_mem;
    ret->capa = BUFINITLEN;

    ret->mark = (fcbmark*)malloc(sizeof(fcbmark));
    if (!ret->mark) goto no_mem;

    return ret;

    no_mem:
        fcb_free(ret);
        return 0;
}


/* free fcbuffer */

void
fcb_free(buf)
    fcbuffer *buf;
{
    fcbmark *m;

    if (buf->ptr) free(buf->ptr);

    m = buf->mark;
    while (m) {
        buf->mark = m->next;
        free(m);
        m = buf->mark;
    }

    free(buf);
}
    

/* number of charactors in buffer */

long
fcb_size(buf)
    fcbuffer *buf;
{
    return FCB_SIZE(buf);
}


/* ------------------------- insert ------------------------ */

static int
incr_bufsize(buf, need)
    fcbuffer *buf;
    long need;
{
    long newcapa;
    fcchar *tmp = 0;

    newcapa = buf->capa;
    while (newcapa < need)
        newcapa = newcapa / 2 * 3;
    tmp = realloc(buf->ptr, fcclen2byte(newcapa));
    if (tmp == 0) return 0;
    buf->ptr = tmp;
    buf->capa = newcapa;

    return 1;
}

static void
mvmark_forins(buf, at, len)
    fcbuffer *buf;
    long at, len;
{
    fcbmark *p;

    for (p = buf->mark; p; p = p->next) {
        if (p->idx >= at)
            break;
    }

    if (!p) return;
    while (p->idx == at) {
        if (p->type->flags & FCBM_L_INSERT) {
            p->idx += len;
            if (p->type->move)
                (*p->type->move)(p, p->idx - len);
        }
        p = p->next;
        if (!p) return;
    }

    if (!p) return;
    while (p->idx > at) {
        p->idx += len;
        if (p->type->move)
            (*p->type->move)(p, p->idx - len);
        p = p->next;
        if (!p) return
    }
}

/*  make gap space

     left   gap  right   capa
    AAAAABB------CCCCCC------
         ^idx
 1  AAAAAooo---BBCCCCCC------   len 3, move B
 2  AAAAAooooooo--BBCCCCCC---   len 7, move C & B

 ................................................

     left   gap  rigth   capa
    AAAAAAA------BBCCCC------
                   ^idx
 3  AAAAAAABBooo---CCCC------   len 3, move B
 4  AAAAAAABBooooooo--CCCC---   len 7, move C & B

 moving C may causes memory allocation.

*/

#define GAP_INC 512

static int
makegap(buf, idx, len)
    fcbuffer *buf;
    long idx, len;
{
    long newlen;

    if (idx <= buf->gap) {
        if (buf->gapsize >= len) {    /* case 1 */
            newlen = buf->gapsize;
        }
        else {                       /* case 2 */
            newlen = len + GAP_INC;
            if (FCB_SIZE(buf) + newlen > buf->capa)
                if (!incr_bufsize(buf)) return 0;
            /* C */
            fcca_cpy(GAPPTR(buf) + newlen,
                     RIGHTPTR(buf),
                     FCB_SIZE(buf) - buf->gap);
        }
        /* B */
        fcca_cpy(buf->ptr + idx + newlen,
                 buf->ptr + idx,
                 buf->gap - idx);
    }
    else {
        if (buf->gapsize >= len) {    /* case 3 */
            newlen = buf->gapsize;
        }
        else {                       /* case 4 */
            newlen = len + GAP_INC;
            if (FCB_SIZE(buf) + newlen > buf->capa)
                if (! incr_bufsize(buf, FCB_SIZE(buf) + newlen)) return 0;
            /* C */
            fcca_cpy(buf->ptr + idx + newlen,
                     buf->ptr + idx + buf->gapsize,
                     FCB_SIZE(buf) - idx);
        }
        /* B */
        fcca_cpy(GAPPTR(buf),
                 RIGHTPTR(buf),
                 idx - buf->gap);
    }
    buf->gap = idx;
    buf->len += newlen - buf->gapsize;
    buf->gapsize = newlen;

    return 1;
}

int
fcb_insert(buf, idx, str)
    fcbuffer *buf;
    long idx;
    fcstring *str;
{
puts("fcb_insert");
    if (idx > FCB_SIZE(buf) || idx < 0) return 0;   /* not >= */

    if (!makegap(buf, idx, str->len)) return 0;
    fcca_cpy(GAPPTR(buf), str->ptr, str->len);
    buf->gap += str->len;
    buf->gapsize -= str->len;

    mvmark_forins(buf, idx, str->len);
        
    return 1;
}


/* -------------------------- delete --------------------------- */

/* delete string in buffer

    delete 3 (len) from ^ (idx).
    "idx" is only abstruct index, NOT real index.

       left     gap    right     capa
    AAAAAABBBCC-----DDEEEEEEEEEE-----

 1  AAAAAABBBCCDD--------EEEEEEE-----
                      ^
 2  AAAAAABBB--------DEEEEEEEEEE-----
             ^
 3  AAAAAA--------CCDDEEEEEEEEEE-----
          ^
*/
    
static void
del_str(buf, idx, len)
    fcbuffer *buf;
    long idx, len;
{
    if (idx >= buf->gap) {          /* case 1 */
        fcca_cpy(GAPPTR(buf),
                 RIGHTPTR(buf),
                 idx - buf->gap);
    }
    else {
        if (idx + len < buf->gap) { /* case 3 */
            fcca_cpy(buf->ptr + idx + len + buf->gapsize,
                     buf->ptr + idx + len,
                     buf->gap - (idx + len));
        }
    }
    buf->gap = idx;
    buf->gapsize += len;
}

static void
mvmark_fordel(buf, from, len)
    fcbuffer *buf;
    long from, len;
{
    fcbmark *p;
    long to, oldi;

    to = from + len;

    for (p = buf->mark; p; p = p->next) {
        if (p->idx > from)   /* not >= */
            break;
    }

    if (!p) return;
    while (p->idx < to) {
        if (p->type->flags & FCBM_REMAIN) {
            oldi = p->idx;
            p->idx = from;
            if (p->type->move)
                (*p->type->move)(p, oldi);
        }
        else {
            if (p->type->delete)
                (*p->type->delete)(p);
            fcb_unmark(buf, p);
        }
        p = p->next;
        if (!p) return;
    }

    if (!p) return;
    while (p->idx >= to) {
        oldi = p->idx;
        p->idx -= len;
        if (p->type->move)
            (*p->type->move)(p, oldi);
        p = p->next;
        if (!p) return;
    }
}

static int
chk_bufsize(buf)
    fcbuffer *buf;
{
    long newcapa;
    fcchar *tmp = 0;

    if (FCB_SIZE(buf) > buf->capa / 8 && buf->capa > BUFINITLEN)
        return;

    newcapa = buf->capa / 3 * 2;
    if (buf->len > newcapa) {
        long rlen;

        rlen = buf->len - RIGHTIDX(buf);
        fcca_cpy(buf->ptr + newcapa - rlen,
                 RIGHTPTR(buf),
                 rlen);
        buf->gaplen = newcapa - buf->gap - rlen;
        buf->len = newcapa;
    }
    tmp = realloc(buf->ptr, newcapa);
    if (!tmp) return 0;
    buf->ptr = tmp;
    buf->capa = newcapa;
}

int
fcb_delete(buf, idx, len)
    fcbuffer *buf;
    long idx, len;
{
puts("fcb_delete");
    if (idx < 0 || idx >= FCB_SIZE(buf)) return 0;
    if (idx + len > FCB_SIZE(buf))
        len = FCB_SIZE(buf) - idx;

    del_str(buf, idx, len);
    mvmark_fordel(buf, idx, len);
    chk_bufsize(buf);

    return 1;
}


/* ----------------------- sub-string ---------------------------- */

static int
copy_substr(buf, idx, len, dest)
    fcbuffer *buf;
    long idx, len;
    fcchar *dest;
{
    if (idx > buf->gap) {
        fcca_cpy(dest,
                 buf->ptr + idx + buf->gapsize,
                 len);
    }
    else {
        if (idx + len > buf->gap) {
            long plen = buf->gap - idx;

            fcca_cpy(dest,
                     buf->ptr + idx,
                     plen);
            fcca_cpy(dest + plen,
                     RIGHTPTR(buf),
                     len - plen);
        }
        else {
            fcca_cpy(dest,
                     buf->ptr + idx,
                     len);
        }
    }
}

/* extract array of fcchar from buffer */

int
fcb_extract_string(buf, idx, len, dest)
    fcbuffer *buf;
    long idx, len;
{
    if (idx < 0 || idx >= FCB_SIZE(buf)) return 0;
    if (idx + len > FCB_SIZE(buf))
        len = FCB_SIZE(buf) - idx;

    copy_substr(buf, idx, len, dest);

    return 1;
}

/* extract fcstring from buffer */

fcstring*
fcb_substr(buf, idx, len)
    fcbuffer *buf;
    long idx, len;
{
    fcstring *ret = 0;
puts("fcb_substr");

    if (idx < 0 || idx >= FCB_SIZE(buf)) return 0;
    if (idx + len > FCB_SIZE(buf))
        len = FCB_SIZE(buf) - idx;

    FCS_NEW_ALLOC(ret, len, 0);
    copy_substr(buf, idx, len, ret->ptr);
    ret->len = len;

    return ret;
}



/* ----------------------------- char ---------------------------- */


/* return char at "idx" */

fcchar
fcb_entry(buf, idx)
    fcbuffer *buf;
    long idx;
{
    if (idx < 0 || idx >= FCB_SIZE(buf)) return FCC_NIL;
    return FCB_ENTRY(buf, idx);
}


/* enter char at "idx" */

int
fcb_enter(buf, char)
    fcbuffer *buf;
    fcchar char;
{
    if (idx < 0 || idx >= FCB_SIZE(buf)) return 0;
    if (idx > buf->gap)
        buf->ptr[idx + buf->gap] = char;
    else
        buf->ptr[idx] = char;
    return 1;
}


/* -------------------------- search ------------------------- */


/* search 'ch' from 'idx' forward (left to right) */

long
fcb_findc(buf, start, ch)
    fcbuffer *buf;
    long idx;
    fcchar ch;
{
    long i;
    fcchar c;

    if (start < 0 || start >= FCB_SIZE(buf)) return FCB_FAIL;

    if (start < buf->gap) {
        for (i = start; i < buf->gap; i++) {
            if (fcc_eq(buf->ptr[i], ch))
                return i;
        }
    }

    i = (start < buf->gap) ? RIGHTIDX(buf) : start + buf->gapsize;
    for (; i < buf->len; i++) {
        if (fcc_eq(buf->ptr[i], ch))
            return i - buf->gapsize;
    }

    return FCB_NOTFOUND;
}


/* search 'ch' from 'idx' backward (start from end) */

long
fcb_rfindc(buf, start, ch)
    fcbuffer *buf;
    long idx;
    fcchar ch;
{
    long i;
    fcchar c;

    if (idx < 0 || idx >= FCB_SIZE(buf)) return FCB_FAIL;

    if (start >= buf->gap) {
        for (i = start + buf->gapsize; i > buf->gap + buf->gapsize; i--) {
            if (fcc_eq(buf->ptr[i], ch))
                return i - buf->gapsize;
        }
    }

    if (buf->gap == 0) return FCB_NOTFOUND;
    i = (start >= buf->gap) ?  buf->gap - 1 : start;
    for (i = idx; i >= 0; i--) {
        if (fcc_eq(buf->ptr[i], ch))
            return i;
    }

    return FCB_NOTFOUND;
}


static long
findnl(buf, start)
    fcbuffer *buf;
    long start;
{
    long i;

    i = fcb_findc(buf, start, FCC_LF);
    if (i == FCB_FAIL)     return FCB_FAIL;
    if (i == FCB_NOTFOUND) return FCB_SIZE(buf);
    return i + 1;
}

static long
rfindnl(buf, start)
    fcbuffer *buf;
    long start;
{
    long i;

    i = fcb_rfindc(buf, start, FCC_LF);
    if (i == FCB_FAIL)      return FCB_FAIL;
    if (i == FCB_NOTFOUND)  return 0;
    return i + 1;
}


/* extract one line which include 'idx' */

fcstring*
fcb_lineat(buf, idx)
    fcbuffer *buf;
    long idx;
{
    long beg, end;

    beg = (idx == 0) ? 0 : rfindnl(buf, idx - 1);
    end = findnl(buf, idx);
    return fcb_substr(buf, beg, end - beg);
}


/* return bare fcchar* of buffer */

fcchar*
fcb_string(buf)
    fcbuffer *buf;
{
    ungap_buffer(buf);
    return buf->ptr;
}


static long
prevline(buf, origin, lines)
    fcbuffer *buf;
    long origin, lines;
{
    long idx, hor, old;

    old = idx = rfindnl(buf, origin);
    hor = origin - idx - 1;

    for (; idx && lines; lines--) {
        old = idx;
        idx = rfindnl(buf, idx);
    }
    old -= idx + 1;
    if (old < 0) old = 0;
    idx += min(old, hor);

    return idx;
}

static long
nextline(buf, origin, lines)
    fcbuffer *buf;
    long origin, lines;
{
    long idx, hor, pre;

    idx = rfindnl(buf, origin);
    hor = origin - idx - 1;

    idx = origin;
    while (idx < buf->len && lines) {
        idx = findnl(buf, idx + 1);
        lines--;
    }
    if (lines) return buf->len;
    pre = findnl(buf, idx + 1);
    if (pre == idx) return idx;
    idx += min(pre - idx - 1, hor);

    return idx;
}

static long
ver_move(buf, origin, lines)
    fcbuffer *buf;
    long origin, lines;
{
    if      (lines == 0) return origin;
    else if (lines > 0)  return prevline(buf, origin, lines);
    else                 return nextline(buf, origin, lines);
}

static long
hor_move(buf, origin, cols)
    fcbuffer *buf;
    long origin, cols;
{
    long i;

    if (cols == 0) return origin;
    if (cols < 0) {
        i = rfindnl(buf, origin);
        return max(i, origin + cols);
    }
    else {
        for (i = origin; i < origin + cols; i++) {
            fcchar ch;

            ch = FCB_CHAR(buf, i);
            if (FCC_ISNL(ch))
                break;
        }
        return i;
    }
}


/* move "m" after "pnt" */

static void
move_mark_after(m, pnt)
    fcbmark *m, *pnt;
{
    m->prev->next = m->next;
    if (m->next)
        m->next->prev = m->prev;
    m->prev = pnt;
    m->next = pnt->next;
    m->prev->next = m;
    if (pnt->next)
        pnt->next->prev = m;
    else
        m->buf->mark_end = m;
}
    

/* insert mark in d-linked list of mark, in fcbuffer
   list is always sorted from small index to big.  */

static void
insert_mark(buf, mark)
    fcbuffer *buf;
    fcbmark *mark;
{
    fcbmark *a, *b;

    for (m = buf->mark; m->next; m = m->next) {
        if (mark->idx < m->next->idx)
            break;
    }
    insert_mark_after(mark, m);

    return;
}

/* attach mark at idx */

fcbmark*
fcb_newmark(buf, idx, type, usr)
    fcbuffer *buf;
    long idx;
    fcbmtype *type;
    void *usr;
{
    fcbmark *p;

    if (idx < 0 || idx > FCB_SIZE(buf)) return 0;
    if (!type) return 0;
                
    p = (fcbmark*)malloc(sizeof(fcbmark));
    if (!p) return 0;

    p->idx = idx;
    p->buf = buf;
    p->prev = 0;
    p->next = 0;
    p->type = type;
    p->usr = usr;

    insert_mark(buf, p);

    return p;
}


static fcbmtype marktype_simple = {
    FCBM_REMAIN | FCBM_L_INSERT,
    0, /* move */
    0, /* delete */
};

fcbmark*
fcb_newmark_simple(buf, idx)
    fcbuffer *buf;
    char *name;
    long idx;
{
    return fcb_newmark(buf, idx, &marktype_simple, 0);
}


void
fcbm_free(mark)
    fcbmark *mark;
{
    free(mark);
}


/* detach mark from fcbuffer */

void
fcb_unmark(buf, mark)
    fcbuffer *buf;
    fcbmark *mark;
{
    if (!mark) return;

    if (mark->prev)
        mark->prev->next = mark->next;
    else
        buf->mark = mark;
    
    if (mark->next)
        mark->next->prev = mark->prev;
    else
        buf->mark_end = mark;
    
    fcbm_free(mark);
}


/* return index of mark */

long
fcbm_index(mark)
    fcbmark *mark;
{
    return mark->idx;
}


/* move mark by relative X pos */

int
fcbm_moveby_x(mark, hor)
    fcbmark *mark;
    long hor;
{
    long i;

    i         = hor ? hor_move(mark->buf, mark->idx, hor) : mark->idx;
    mark->idx = ver ? ver_move(mark->buf, i, ver) : i;
    return 1;
}


/* move mark by relative Y pos */

int
fcbm_moveby_y(mark, ver)
    fcbmark *mark;
    long ver;
{
    long i;

    i         = hor ? hor_move(mark->buf, mark->idx, hor) : mark->idx;
    mark->idx = ver ? ver_move(mark->buf, i, ver) : i;
    return 1;
}


/* move mark by relative X-Y pos */

int
fcbm_moveby_xy(mark, hor, ver)
    fcbmark *mark;
    long hor, ver;
{
    fcbm_moveby_y(mark->buf, mark->idx, ver);
    fcbm_moveby_x(mark->buf, mark->idx, hor);
    return 1;
}


/* move mark to absolute X-Y pos */

int
fcbm_moveto_xy(mark, hor, ver)
    fcbmark *mark;
    long hor, ver;
{
    return 0;
}


/* move mark "len" chars */

int
fcbm_moveby_i(mark, by)
    fcbmark *mark;
    long by;
{
    fcbmark *m;
    long to;

    to = mark->idx + by;
    if (to < 0 || to > FCB_SIZE(mark->buf)) return 0;

    if (by > 0) {
        for (m = mark; m->next; m = m->next) {
            if (m->next->idx > to)
                break;
        }
    }
    else {
        for (m = mark; m->prev; m = m->prev) {
            if (m->prev->idx < to) {
                m = m->prev;
                break;
            }
        }
    }
    if (m != mark)
        insert_mark_after(mark, m);
    mark->idx = to;
}


/* move mark to "to" */

int
fcbm_moveto_i(mark, to)
    fcbmark *mark;
    long to;
{
    fcbmark *m;

    if (to < 0 || to > FCB_SIZE(mark->buf)) return 0;

    return 1;
}


/* insert string at mark */

int
fcbm_insert(mark, str)
    fcbmark *mark;
    fcstring *str;
{
    return fcb_insert(mark->buf, mark->idx, str);
}


/* delete string at mark */

int
fcbm_delete(mark, len)
    fcbmark *mark;
    long len;
{
    if (len < 0) {
        len *= -1;
        len = max(mark->idx - len, 0);
        return fcb_delete(mark->buf, mark->idx - len, len);
    }
    else
        return fcb_delete(mark->buf, mark->idx, len);
}
