/* vi:set ts=4 sw=4:
==============================================================

  fcstring version 0.0.0

  Copyright(c) 1999 Minero Aoki
  aamine@dp.u-netsurf.ne.jp

  this program is Free Software.

==============================================================
*/
/*
source code for fchash is derived from st.h by Peter Moore

This is a general purpose hash table package
written by Peter Moore @ UCB.

@(#) st.h 5.1 89/12/14

*/

# ifndef FCSTRING_H
# define FCSTRING_H


/* temp */
#ifndef HAVE_PROTOTYPES
#define HAVE_PROTOTYPES
#endif

#undef _
#ifdef HAVE_PROTOTYPES
#  define _(args) args
#else
#  define _(args) ()
#endif


typedef  unsigned long  fcchar;
typedef  unsigned int   fcctype;

typedef struct fcstring fcstring;
typedef struct fcbuffer fcbuffer;
typedef struct fcbmark fcbmark;
typedef struct fcbmtype fcbmtype;

typedef struct fchash fchash;


struct fcstring
{
    fcchar *ptr;
    long    len;
};


struct fcbuffer
{
    fcchar  *ptr;
    long     len;
    long     capa;
    long     gap;
    long     gapsize;

    fcbmark *mark;
    fcbmark *mark_end;
};


struct fcbmark
{
    fcbuffer *buf;
    long      idx;
    char     *name;

    fcbmark  *prev;
    fcbmark  *next;

    fcbmtype *type;
    void *usr;
};


struct fcbmtype
{
    unsigned int flags;
    void (*move)(fcbmark *mark, long oldi);
    void (*delete)(fcbmark *mark);
};


/* consts ------------------------------------------------ */


# define  FCC_TYPE_SPETIAL     0x0000U
# define  FCC_TYPE_EN          0x2842U
# define  FCC_TYPE_JA          0x2442U


# define  FCC_NIL              0x00000000UL
# define  FCC_CRLF             0x0000000aUL

# define  FCC_NUL              0x28422000UL
# define  FCC_SOH              (FCC_NUL + 1)
# define  FCC_STX              (FCC_NUL + 2)
# define  FCC_ETX              (FCC_NUL + 3)
# define  FCC_EOT              (FCC_NUL + 4)
# define  FCC_ENQ              (FCC_NUL + 5)
# define  FCC_ACK              (FCC_NUL + 6)
# define  FCC_BEL              (FCC_NUL + 7)
# define  FCC_BS               (FCC_NUL + 8)
# define  FCC_HT               (FCC_NUL + 9)
# define  FCC_LF               (FCC_NUL + 10)
# define  FCC_VT               (FCC_NUL + 11)
# define  FCC_FF               (FCC_NUL + 12)
# define  FCC_CR               (FCC_NUL + 13)
# define  FCC_SO               (FCC_NUL + 14)
# define  FCC_SI               (FCC_NUL + 15)
# define  FCC_DLE              (FCC_NUL + 16)
# define  FCC_DC1              (FCC_NUL + 17)
# define  FCC_DC2              (FCC_NUL + 18)
# define  FCC_DC3              (FCC_NUL + 19)
# define  FCC_DC4              (FCC_NUL + 20)
# define  FCC_NAK              (FCC_NUL + 21)
# define  FCC_SYN              (FCC_NUL + 22)
# define  FCC_ETB              (FCC_NUL + 23)
# define  FCC_CAN              (FCC_NUL + 24)
# define  FCC_EM               (FCC_NUL + 25)
# define  FCC_SUB              (FCC_NUL + 26)
# define  FCC_ESC              (FCC_NUL + 27)
# define  FCC_IS4              (FCC_NUL + 28)
# define  FCC_IS3              (FCC_NUL + 29)
# define  FCC_IS2              (FCC_NUL + 30)
# define  FCC_IS1              (FCC_NUL + 31)
# define  FCC_SPACE            (FCC_NUL + 32)
# define  FCC_EXCLAM           (FCC_NUL + 33)
# define  FCC_QUOTEDBL         (FCC_NUL + 34)
# define  FCC_NUMBERSIGN       (FCC_NUL + 35)
# define  FCC_DOLLER           (FCC_NUL + 36)
# define  FCC_PERCENT          (FCC_NUL + 37)
# define  FCC_AMPERSAND        (FCC_NUL + 38)
# define  FCC_APOSTROPHE       (FCC_NUL + 39)
# define  FCC_PARENLEFT        (FCC_NUL + 40)
# define  FCC_PARENRIGHT       (FCC_NUL + 41)
# define  FCC_ASTERISK         (FCC_NUL + 42)
# define  FCC_PLUS             (FCC_NUL + 43)
# define  FCC_COMMA            (FCC_NUL + 44)
# define  FCC_MINUS            (FCC_NUL + 45)
# define  FCC_PERIOD           (FCC_NUL + 46)
# define  FCC_SLASH            (FCC_NUL + 47)
# define  FCC_0                (FCC_NUL + 48)
# define  FCC_1                (FCC_NUL + 49)
# define  FCC_2                (FCC_NUL + 50)
# define  FCC_3                (FCC_NUL + 51)
# define  FCC_4                (FCC_NUL + 52)
# define  FCC_5                (FCC_NUL + 53)
# define  FCC_6                (FCC_NUL + 54)
# define  FCC_7                (FCC_NUL + 55)
# define  FCC_8                (FCC_NUL + 56)
# define  FCC_9                (FCC_NUL + 57)
# define  FCC_COLON            (FCC_NUL + 58)
# define  FCC_SEMICOLON        (FCC_NUL + 59)
# define  FCC_LESS             (FCC_NUL + 60)
# define  FCC_EQUAL            (FCC_NUL + 61)
# define  FCC_GREATER          (FCC_NUL + 62)
# define  FCC_QUESTION         (FCC_NUL + 63)
# define  FCC_AT               (FCC_NUL + 64)
# define  FCC_A                (FCC_NUL + 65)
# define  FCC_B                (FCC_NUL + 66)
# define  FCC_C                (FCC_NUL + 67)
# define  FCC_D                (FCC_NUL + 68)
# define  FCC_E                (FCC_NUL + 69)
# define  FCC_F                (FCC_NUL + 70)
# define  FCC_G                (FCC_NUL + 71)
# define  FCC_H                (FCC_NUL + 72)
# define  FCC_I                (FCC_NUL + 73)
# define  FCC_J                (FCC_NUL + 74)
# define  FCC_K                (FCC_NUL + 75)
# define  FCC_L                (FCC_NUL + 76)
# define  FCC_M                (FCC_NUL + 77)
# define  FCC_N                (FCC_NUL + 78)
# define  FCC_O                (FCC_NUL + 79)
# define  FCC_P                (FCC_NUL + 80)
# define  FCC_Q                (FCC_NUL + 81)
# define  FCC_R                (FCC_NUL + 82)
# define  FCC_S                (FCC_NUL + 83)
# define  FCC_T                (FCC_NUL + 84)
# define  FCC_U                (FCC_NUL + 85)
# define  FCC_V                (FCC_NUL + 86)
# define  FCC_W                (FCC_NUL + 87)
# define  FCC_X                (FCC_NUL + 88)
# define  FCC_Y                (FCC_NUL + 89)
# define  FCC_Z                (FCC_NUL + 90)
# define  FCC_BRACKETLEFT      (FCC_NUL + 91)
# define  FCC_BACKSLASH        (FCC_NUL + 92)
# define  FCC_BRACKETRIGHT     (FCC_NUL + 93)
# define  FCC_ASCIICIRCUM      (FCC_NUL + 94)
# define  FCC_UNDERSCORE       (FCC_NUL + 95)
# define  FCC_QUOTELEFT        (FCC_NUL + 96)
# define  FCC_a                (FCC_NUL + 97)
# define  FCC_b                (FCC_NUL + 98)
# define  FCC_c                (FCC_NUL + 99)
# define  FCC_d                (FCC_NUL + 100)
# define  FCC_e                (FCC_NUL + 101)
# define  FCC_f                (FCC_NUL + 102)
# define  FCC_g                (FCC_NUL + 103)
# define  FCC_h                (FCC_NUL + 104)
# define  FCC_i                (FCC_NUL + 105)
# define  FCC_j                (FCC_NUL + 106)
# define  FCC_k                (FCC_NUL + 107)
# define  FCC_l                (FCC_NUL + 108)
# define  FCC_m                (FCC_NUL + 109)
# define  FCC_n                (FCC_NUL + 110)
# define  FCC_o                (FCC_NUL + 111)
# define  FCC_p                (FCC_NUL + 112)
# define  FCC_q                (FCC_NUL + 113)
# define  FCC_r                (FCC_NUL + 114)
# define  FCC_s                (FCC_NUL + 115)
# define  FCC_t                (FCC_NUL + 116)
# define  FCC_u                (FCC_NUL + 117)
# define  FCC_v                (FCC_NUL + 118)
# define  FCC_w                (FCC_NUL + 119)
# define  FCC_x                (FCC_NUL + 120)
# define  FCC_y                (FCC_NUL + 121)
# define  FCC_z                (FCC_NUL + 122)
# define  FCC_BRACELEFT        (FCC_NUL + 123)
# define  FCC_BAR              (FCC_NUL + 124)
# define  FCC_BRACERIGHT       (FCC_NUL + 125)
# define  FCC_ASCIITILDE       (FCC_NUL + 126)
# define  FCC_DEL              (FCC_NUL + 127)



/* class fcchar ------------------------------------------ */


# define  FCC_PAD              0
# define  FCC_TYPEMASK         0x0000ffffUL

#define        fcc_type(ch)         (((ch) >> 16) & FCC_TYPEMASK)
#define        fcc_typetest(ch, t)  (fcc_type(ch) == (t))

#define        fcc_froma(b)  (FCC_NUL+(b))

#define        fcc_from2b(type, b1, b2)  (FCC_NIL+((type)<<16)+((b1)<<8)+(b2))
#define        fcc_from1b(type, b)  fcc_from2b(type, FCC_PAD, b)


#define        fcc_a1(ch)  ((char)(((ch) >> 8)  & 0x000000ffUL))
#define        fcc_a2(ch)  ((char)( (ch)        & 0x000000ffUL))

#define        fcc_settype(ch, type)  ((ch) | (type << 16))
#define        fcc_seta1(ch, a)       ((ch) | ((a) << 8))
#define        fcc_seta2(ch, a)       ((ch) | (a))


#define        fcc_eq(a,b)    ((a) == (b))
#define        fcc_gt(a,b)    ((a) >  (b))
#define        fcc_le(a,b)    ((a) <  (b))
#define        fcc_gteq(a,b)  ((a) >= (b))
#define        fcc_leeq(a,b)  ((a) <= (b))

#define        fcc_aeq(a,b)     ((a) == fcc_froma(FCC_TYPE_EN, b))
#define        fcc_agt(a,b)     ((a) >  fcc_froma(FCC_TYPE_EN, b))
#define        fcc_ale(a,b)     ((a) <  fcc_froma(FCC_TYPE_EN, b))
#define        fcc_agteq(a,b)   ((a) >= fcc_froma(FCC_TYPE_EN, b))
#define        fcc_aleeq(a,b)   ((a) <= fcc_froma(FCC_TYPE_EN, b))

#define        fcc_issb(ch)  (fcc_a1(ch) == FCC_PAD)
#define        fcc_ismb(ch)  (fcc_a1(ch) != FCC_PAD)


#define        FCC_ISNL(ch)  ((ch) == FCC_LF || \
                              (ch) == FCC_CRLF || \
                              (ch) == FCC_CR)

#define        FCC_ISALPHA(ch)  ((((ch) == FCC_a) && ((ch) <= FCC_z)) || \
                                 (((ch) >= FCC_A) && ((ch) <= FCC_Z)))

#define        FCC_ISLOWER(ch)  ((((ch) >= FCC_a) && (ch <= FCC_z)))

#define        FCC_ISUPPER(ch)  ((((ch) >= FCC_A) && (ch <= FCC_Z)))

#define        FCC_ISDIGIT(ch)  ((((ch) >= FCC_0) && (ch <= FCC_9)))

#define        FCC_ISALNUM(ch)  (FCC_ISALPHA(ch) || FCC_ISDIGIT(ch))

#define        FCC_ISCNTRL(ch)  ((((ch) > FCC_NUL) && ((ch) <= FCC_IS1)))

#define        FCC_ISSPACE(ch)  (((ch) == FCC_SPACE) || \
                                 ((ch) == FCC_LF)    || \
                                 ((ch) == FCC_HT)    || \
                                 ((ch) == FCC_CRLF)  || \
                                 ((ch) == FCC_CR)    || \
                                 ((ch) == FCC_VT)    || \
                                 ((ch) == FCC_FF))

#define        FCC_ISGRAPH(ch)  (((ch) > FCC_SPACE) && \
                                 ((ch) <= FCC_ASCIITILDE))

#define        FCC_ISPRINT(ch)  (((ch) >= FCC_SPACE) && \
                                 ((ch) <= FCC_ASCIITILDE))

#define        FCC_ISPUNCT(ch)  (FCC_ISGRAPH(ch) && !FCC_ISALNUM(ch))

#define        FCC_ISXDIGIT(ch)  (FCC_ISDIGIT(ch) || \
                                 (((ch) >= FCC_a) && ((ch) <= FCC_f)) || \
                                 (((ch) >= FCC_A) && ((ch) <= FCC_F)))

#define        FCC_ISBLANK(ch)  (((ch) == FCC_SPACE) || ((ch) == FCC_HT))


extern int     fcc_isnl _((fcchar ch));
extern int     fcc_isalpha _((fcchar ch));
extern int     fcc_islower _((fcchar ch));
extern int     fcc_isupper _((fcchar ch));
extern int     fcc_isalnum _((fcchar ch));
extern int     fcc_isdigit _((fcchar ch));
extern int     fcc_isprint _((fcchar ch));
extern int     fcc_ispunct _((fcchar ch));
extern int     fcc_iscntrl _((fcchar ch));
extern int     fcc_isspace _((fcchar ch));
extern int     fcc_isgraph _((fcchar ch));
extern int     fcc_isblank _((fcchar ch));
extern int     fcc_isxdigit _((fcchar ch));


#define  FCC_TOLOWER(ch)  (FCC_ISUPPER(ch) ? ((ch) + 0x20UL) : (ch))
#define  FCC_TOUPPER(ch)  (FCC_ISLOWER(ch) ? ((ch) - 0x20UL) : (ch))

extern fcchar  fcc_tolower _((fcchar ch));
extern fcchar  fcc_toupper _((fcchar ch));



/* class fcstring ---------------------------------------- */

extern fcstring* fcs_new _((void));
extern void fcs_free _((fcstring *str));

#define  fcs_length(str) ((str)->len)
#define  fcs_size(str)   ((str)->len)
#define  fcs_empty(str)  ((str)->len == 0)
#define  fcs_string(str) ((str)->ptr)

extern int fcs_equal _((fcstring *stra, fcstring *strb));
extern int fcs_cequal _((fcstring *stra, fcstring *strb));
extern int fcs_compare _((fcstring *stra, fcstring *strb));
extern int fcs_ccompare _((fcstring *stra, fcstring *strb));

extern fcchar fcs_entry _((fcstring *str, long idx));
extern fcstring* fcs_substr _((fcstring *str, long idx, long len));
extern fcstring* fcs_dup _((fcstring *src));
extern fcstring* fcs_plus _((fcstring *befstr, fcstring *afstr));
extern fcstring* fcs_times _((fcstring *str, long times));

#define    fcs_prepend(dest, src)  fcs_insert(dest, 0, src)
extern int fcs_concat _((fcstring *dest, fcstring *src));
#define    fcs_append(dest, src)  fcs_concat(dest, src)
extern int fcs_replace _((fcstring *dest, fcstring *src));
extern int fcs_clear _((fcstring *str));
extern int fcs_set _((fcstring *dest, long idx, fcstring *src));
extern int fcs_enter _((fcstring *str, long idx, fcchar chr));
extern int fcs_insert _((fcstring *dest, long idx, fcstring *src));
extern int fcs_insertc _((fcstring *str, long idx, fcchar ch));
extern int fcs_delete _((fcstring *str, long idx, long len));
extern int fcs_deletec _((fcstring *str, long idx));
extern int fcs_chop _((fcstring *str));
extern int fcs_chomp _((fcstring *str));
extern int fcs_strip _((fcstring *str));
#define    fcs_cuth(str, len)  fcs_delete(str, 0, len)
extern int fcs_cutt _((fcstring *str, long len));
extern int fcs_toupper _((fcstring *str));
extern int fcs_tolower _((fcstring *str));

#define  FCS_NOTFOUND  (-1)
extern long fcs_count _((fcstring *str, fcchar ch));
extern long fcs_findc _((fcstring *str, fcchar ch));
extern long fcs_rfindc _((fcstring *str, fcchar ch));
extern long fcs_find _((fcstring *str, fcstring *targ));
extern long fcs_rfind _((fcstring *str, fcstring *targ));

extern char* fcs_to_iso2022 _((fcstring *str));
extern fcstring* fcs_from_out _((fcchar *arr, long len));
extern fcstring* fcs_from_out_dup _((fcchar *arr, long len));
extern fcstring* fcs_from_iso2022 _((char *arr, size_t arrsize));
extern fcstring* fcs_from_eucjp _((char *arr, size_t arrsize));
extern fcstring* fcs_from_sjis _((char *arr, size_t arrsize));


/* class fcbuffer ---------------------------------------- */


#define  FCB_SIZE(buf)  ((buf)->len - (buf)->gapsize)
#define  FCB_LEN(buf)   FCB_SIZE(buf)
#define  FCB_ENTRY(buf, i) \
    ((i) < (buf)->gap ? (buf)->ptr[i] : (buf)->ptr[(i) + (buf)->gap])

extern fcbuffer *fcb_new _((void));
extern void fcb_free _((fcbuffer *buf));

extern long fcb_size _((fcbuffer *buf));
#define     fcb_length(buf)  fcb_size(buf)

extern int fcb_insert _((fcbuffer *buf, long idx, fcstring *str));
extern int fcb_delete _((fcbuffer *buf, long idx, long len));
/*
extern int fcb_overwrite _((fcbuffer *buf, long idx, fcstring *str));
*/

extern fcstring *fcb_substr _((fcbuffer *buf, long idx, long len));
extern fcstring *fcb_lineat _((fcbuffer *buf, long idx));
extern fcchar fcb_entry _((fcbuffer *buf, long idx));
extern int fcb_enter _((fcbuffer *buf, fcchar char));

#define  FCB_NOTFOUND  (-1)
#define  FCB_FAIL      (-2)
extern long fcb_findc _((fcbuffer *buf, long idx, fcchar ch));
extern long fcb_rfindc _((fcbuffer *buf, long idx, fcchar ch));

extern fcchar *fcb_string _((fcbuffer *buf));


/* fcbmark type flags */

#define FCBM_REMAIN    (1U << 0)
#define FCBM_DELETE    0
#define FCBM_L_INSERT  (1U << 1)
#define FCBM_R_INSERT  0
    
extern fcbmark *fcb_newmark _((fcbuffer *buf, long idx,
                               fcbmtype *type, void *usr));
extern fcbmark *fcb_newmark_simple _((fcbuffer *buf, long idx));
extern void fcbm_free _((fcbmark *mark));
extern void fcb_unmark _((fcbuffer *buf, fcbmark *mark));

extern long fcbm_index _((fcbmark *mark));
extern int fcbm_moveby_i _((fcbmark *mark, long len));
extern int fcbm_moveto_i _((fcbmark *mark, long to));
extern int fcbm_moveby_xy _((fcbmark *mark, long hor, long ver));
extern int fcbm_moveto_xy _((fcbmark *mark, long hor, long ver));
extern int fcbm_insert _((fcbmark *mark, fcstring *str));
extern int fcbm_delete _((fcbmark *mark, long len));


# endif  /* FCSTRING_H */
